# This is a Julia file, tested with Julia v1.5.2-1
# Julia can be downloaded from https://julialang.org/downloads/
# I recommend using JuliaPro instead, which also gives you an IDE.
# Download JuliaPro from https://juliacomputing.com/products/juliapro.html
# This file needs the Distributions package. Install it by running the following:
#    Pkg.add("Distributions")
# This command is Julia code, and should be run as normal Julia code in the REPL,
# or in the Julia IDE by uncommenting that line and pressing either
# CTRL+Enter (on Linux/Windows)  or CMD+Return (on a Mac).

# This is an implementation of an internal DSL for probabilistic programming.
# The key difference with other probabilistic programming languages is that
# our observe(D,I) operation takes an interval I,
# whereas in other languages observe(D,x) takes a single point x.
# Our interval I may have finite width or infintesimally small width, so we can
# still perform conditioning on events of measure zero.
# This fixes paradoxical behaviour. In particular, programs with Intervals
# behave in a consistent manner when we use different units (e.g. meters or centimeters)
# and even when we peform non-linear parameter transformations.

# Section 3 of the paper contains three motivating examples. These examples are
# at the end of this file. The key point is that performing a parameter transformation
# does not change the results.
# This is probably not very exciting to test, but it is an important well-behavedness
# criterion that the standard implementation violates.
# We also make changes to some of the examples to simulate what other implementations do.
# In our language the programming error is visible in the source program, for instance
# as a unit error, where the midpoint of the interval is in centimeters but the width
# of the interval is in meters.

import Base.+, Base.-, Base.*, Base./, Base.&, Base.|, Base.!, Base.promote_rule, Base.convert, Base.rand
using Random, Distributions
import Distributions.pdf, Distributions.cdf



### Infinitesimals
# We define a data type of infinitesimal numbers, and arithmetic operations on them

struct Infinitesimal <: Number
    # this represents the number r*epsilon^n
    r::Float64
    n::Int32
end

function (+)(x::Infinitesimal, y::Infinitesimal)
    if x.n < y.n
        x
    elseif x.n > y.n
        y
    else
        Infinitesimal(x.r+y.r, x.n)
    end
end

function(-)(x::Infinitesimal, y::Infinitesimal)
    x + Infinitesimal(-y.r,y.n)
end

function (*)(x::Infinitesimal, y::Infinitesimal)
    Infinitesimal(x.r*y.r, x.n+y.n)
end

function (/)(x::Infinitesimal, y::Infinitesimal)
    Infinitesimal(x.r/y.r, x.n-y.n)
end

# conversion and promotion rules allow us to mix infinitesimals with floats and ints
convert(::Type{Infinitesimal}, x::Float64) = Infinitesimal(x,0)
convert(::Type{Infinitesimal}, x::Int64) = Infinitesimal(x,0)

promote_rule(::Type{Float64}, ::Type{Infinitesimal}) = Infinitesimal
promote_rule(::Type{Infinitesimal}, ::Type{Float64}) = Infinitesimal
promote_rule(::Type{Int64}, ::Type{Infinitesimal}) = Infinitesimal
promote_rule(::Type{Infinitesimal}, ::Type{Int64}) = Infinitesimal

eps = Infinitesimal(1.0,1)



### Intervals

struct Interval
    mid
    width # the width can be infinitesimal
end

# the probability that a value sampled from d is equal to x
function prob(d::Distribution{Univariate,Discrete}, x)
    pdf(d,x)
end

function interval_cdf(d, mid, width::Infinitesimal)
    if width.n == 0
        cdf(d,mid+width.r/2) - cdf(d,mid-width.r/2)
    else
        pdf(d,mid)*width
    end
end

function interval_cdf(d, mid, width)
    interval_cdf(d,mid,Infinitesimal(width,0))
end

# the probability that a value sampled form d lies in the interval x
# this works for finite width intervals as well as infinitesimal width intervals
function prob(d::Distribution{Univariate,Continuous}, x::Interval)
    interval_cdf(d,x.mid,x.width)
end



### Parameter transformations

struct Transformation
    f # the main function, assumed to be monotone

    # auxilary functions we need to perform parameter transformations with f
    fprime # derivative of f
    finv # inverse of f
    finvprime # deriviative of inverse of f
end

function inverse(t::Transformation)
    Transformation(t.finv, t.finvprime, t.f, t.fprime)
end

constant(x) = y -> x
multiply(x) = y -> x*y

# this can be used to change units, e.g. mulT(100) for meters -> centimeters
mulT(a) = Transformation(multiply(a), constant(a), multiply(1/a), constant(1/a))

# the exp/log parameter transformation can be used to change from
# decibels to energy density
expT = Transformation(exp, exp, log, x -> 1/x)

# this function applies t to an interval: [a,b] -> [f(a), f(b)]
# but since our intervals are stored as midpoint-width pairs it's a bit more involved
function applyTtoInterval(t, mid, width)
    left = t.f(mid - width / 2)
    right = t.f(mid + width / 2)
    Interval((left + right) / 2, right - left)
end

function applyTtoInterval(t, mid, width::Infinitesimal)
    if width.n == 0
        applyTtoInterval(t, mid, width.r)
    else
        # if the width is an infinitesimal number we need to use the derivative of t
        Interval(t.f(mid), t.fprime(mid) * width)
    end
end

# syntactic sugar so that we can use function call syntax when applying a
# parameter transformation to an interval
function (t::Transformation)(x::Interval)
    applyTtoInterval(t,x.mid,x.width)
end

# a data type to represent transformed probability distributions
# we define random sampling, pdf, and cdf
struct Transformed <: ContinuousUnivariateDistribution
    original::ContinuousUnivariateDistribution
    t::Transformation
end

function rand(d::Transformed)
    d.t.f(rand(d.original))
end

function pdf(d::Transformed, x::Real)
    pdf(d.original, d.t.finv(x))*d.t.finvprime(x)
end

function cdf(d::Transformed, x::Real)
    cdf(d.original, d.t.finv(x))
end

# syntactic sugar for applying a parameter transformation to a distribution
function (t::Transformation)(d::ContinuousUnivariateDistribution)
    Transformed(d,t)
end


### Importance sampling

# this implementation of importance sampling is entirely standard, except that
# we use infinitesimal numbers to represent the likelihood weight

weight = [Infinitesimal(1.0,0)] # use a one element array because global variables
                                # cannot be mutated from within functions

function observe(D,I)
    weight[1] *= prob(D,I)
end

function importance(trials, program)
    function run()
        weight[1] = Infinitesimal(1.0,0)
        ret = program()
        return (ret,weight[1])
    end

    results = [run() for i in 1:trials]

    total_weight = sum(w for (r,w) in results)
    total_weighted_ret = sum(r*w for (r,w) in results)

    return (total_weighted_ret/total_weight).r
end

### Examples

# first come the 3 paradoxical examples from section 3

# the first example from section 3, calculating bmi

function example1_bmi_m()
    h = rand(Normal(1.70, 0.2))
    w = rand(Normal(70, 30))
    if rand(Bernoulli(0.5))
        observe(Normal(2.0,0.1), Interval(h,eps))
    else
        observe(Normal(90,5), Interval(w,eps))
    end
    return w / h^2
end

# the result of this should (approximately) match the result of the next program
importance(1000000,example1_bmi_m)

function example1_bmi_cm()
    h = rand(Normal(170, 20)) # switch from meters to centimeters
    w = rand(Normal(70, 30))
    if rand(Bernoulli(0.5))
        observe(Normal(200,10), Interval(h,100*eps)) # factor of 100 required here to keep the units consistent
    else
        observe(Normal(90,5), Interval(w,eps))
    end
    return w / (0.01*h)^2
end

# the result of this should (approximately) match the result of the previous program
importance(1000000,example1_bmi_cm)

# this is an illustration of what other probabilistic programming languages compute
# here we forgot to include the unit conversion in the width of the interval
# that is, we use eps instead of 100*eps
# in other probabilistic programming languages it is not even possible to include this factor
# so in those languages it is implicitly 1, and the unit error is not visible in the
# source program.

function example1_bmi_bad()
    h = rand(Normal(170, 20))
    w = rand(Normal(70, 30))
    if rand(Bernoulli(0.5))
        observe(Normal(200,10), Interval(h,eps)) # if we forget the 100 factor here we get a different result
    else
        observe(Normal(90,5), Interval(w,eps))
    end
    return w / (0.01*h)^2
end

# this result does not match the previous two results
importance(1000000,example1_bmi_bad)

# the second example of section 3, expressed in our language
# unlike in Anglican, here the m (meters) and cm (centimeters) programs
# return consistent results.

function example2_m()
    h = rand(Normal(1.7,0.5))
    if rand(Bernoulli(0.5))
        observe(Normal(2.0,0.1), Interval(h,eps))
    end
    return h
end

# should be approximately 1.7
importance(1000000,example2_m)

function example2_cm()
    h = rand(Normal(170,50))
    if rand(Bernoulli(0.5))
        observe(Normal(200,10), Interval(h,100*eps))
    end
    return h
end

# should be approximately 170
importance(1000000,example2_cm)

# the same example, but now with an interval of finite width
# decreasing the number 80 to zero should match the previous results
function example2_finite_width()
    h = rand(Normal(170,50))
    if rand(Bernoulli(0.5))
        observe(Normal(200,10), Interval(h,80)) # play with the number 80 here
    end
    return h
end

importance(1000000,example2_finite_width)

# the third example from section 3, non-linear parameter transformation

function example3_decibel()
    x = rand(Normal(10,5))
    observe(Normal(15,5),Interval(x,eps))
    return exp(x)
end

# the result of this should (approximately) match the result of the next program
importance(1000000,example3_decibel)

function example3_energy_density()
    exp_x = rand(LogNormal(10,5))
    observe(LogNormal(15,5),Interval(exp_x,exp_x*eps))
    return exp_x
end

# the result of this should (approximately) match the result of the previous program
importance(1000000,example3_energy_density)

function example3_energy_density_bad()
    exp_x = rand(LogNormal(10,5))
    observe(LogNormal(15,5),Interval(exp_x,eps))
    return exp_x
end

# the result of this does not match the result of the previous two programs
# because the width of the interval in the program was incorrectly transformed.
# instead of exp_x*eps we have eps.
# in other probabilistic programming languages it is not even possible to include
# this factor exp_x, so in those languages it is implicitly 1, and the error is
# not visible in the source program.
importance(1000000,example3_energy_density_bad)


# this shows the use of explicit parameter transformations to perform the conversion

# the output of all four programs should be (approximately) the same

# the original program
function example3_transform1()
    x = rand(Normal(10,5))
    observe(Normal(15,5),Interval(x,eps))
    return exp(x)
end

importance(1000000,example3_transform1)

# we first transform the input to the observe call
function example3_transform2()
    x = rand(Normal(10,5))
    observe(expT(Normal(15,5)), expT(Interval(x,eps)))
    return exp(x)
end

importance(1000000,example3_transform2)

# we now transform the variable x to exp(x)
function example3_transform3()
    exp_x = rand(expT(Normal(10,5)))
    observe(expT(Normal(15,5)), expT(Interval(log(exp_x),eps)))
    return exp_x
end

importance(1000000,example3_transform3)

# we simplify expT(Interval(log(exp_x),eps)) to Interval(exp_x,exp_x*eps)
# thus obtaining the correctly transformed program
function example3_transform4()
    exp_x = rand(LogNormal(10,5))
    observe(LogNormal(15,5),Interval(exp_x,exp_x*eps))
    return exp_x
end

importance(1000000,example3_transform4)
